import os
import sys
import shutil
import tempfile
import contextlib
from StringIO import StringIO

import pyblish
import pyblish.cli
import pyblish.plugin

# Setup
HOST = 'python'
FAMILY = 'test.family'

REGISTERED = pyblish.plugin.registered_paths()
PACKAGEPATH = pyblish.lib.main_package_path()
ENVIRONMENT = os.environ.get("PYBLISHPLUGINPATH", "")
PLUGINPATH = os.path.join(PACKAGEPATH, '..', 'tests', 'plugins')


def setup():
    """Disable default plugins and only use test plugins"""
    pyblish.plugin.deregister_all_paths()


def setup_empty():
    """Disable all plug-ins"""
    setup()
    pyblish.plugin.deregister_all_plugins()
    pyblish.plugin.deregister_all_paths()
    pyblish.plugin.deregister_all_hosts()
    pyblish.plugin.deregister_all_callbacks()


def teardown():
    """Restore previously REGISTERED paths"""

    pyblish.plugin.deregister_all_paths()
    for path in REGISTERED:
        pyblish.plugin.register_plugin_path(path)

    os.environ["PYBLISHPLUGINPATH"] = ENVIRONMENT
    pyblish.api.deregister_all_plugins()
    pyblish.api.deregister_all_hosts()
    pyblish.api.deregister_test()
    pyblish.api.__init__()


@contextlib.contextmanager
def captured_stdout():
    """Temporarily reassign stdout to a local variable"""
    try:
        sys.stdout = StringIO()
        yield sys.stdout
    finally:
        sys.stdout = sys.__stdout__


@contextlib.contextmanager
def captured_stderr():
    """Temporarily reassign stderr to a local variable"""
    try:
        sys.stderr = StringIO()
        yield sys.stderr
    finally:
        sys.stderr = sys.__stderr__


@contextlib.contextmanager
def tempdir():
    """Provide path to temporary directory"""
    try:
        tempdir = tempfile.mkdtemp()
        yield tempdir
    finally:
        shutil.rmtree(tempdir)




import getpass
import pyblish.api


class CollectCurrentUser(pyblish.api.ContextPlugin):
    """Inject the currently logged on user into the Context"""

    order = pyblish.api.CollectorOrder

    def process(self, context):
        context.data['user'] = getpass.getuser()



"""Conveinence function for Pyblish

Attributes:
    TAB: Number of spaces for a tab
    LOG_TEMPATE: Template used for logging coming from
        plug-ins
    SCREEN_WIDTH: Default width at which logging and printing
        will (attempt to) restrain to.
    logging_handlers: Record of handlers at the start of
        importing this module. This module will modify the
        currently handlers and restore then once finished.
    log: Current logger
    intro_message: Message printed upon initiating a publish.

"""

from __future__ import absolute_import

# Standard library
import logging
import warnings

# Local library
from . import logic, plugin

log = logging.getLogger("pyblish.util")


def publish(context=None, plugins=None, **kwargs):
    """Publish everything

    This function will process all available plugins of the
    currently running host, publishing anything picked up
    during selection.

    Arguments:
        context (plugin.Context): Optional Context,
            defaults to creating a new context
        plugins (list): (Optional) Plug-ins to include,
            defaults to discover()

    Usage:
        >> context = plugin.Context()
        >> publish(context)  # Pass..
        >> context = publish()  # ..or receive a new

    """

    # Must check against None, as the
    # Context may come in empty.
    if context is None:
        context = plugin.Context()

    if plugins is None:
        plugins = plugin.discover()

    # Do not consider inactive plug-ins
    plugins = list(p for p in plugins if p.active)

    test = logic.registered_test()
    state = {
        "nextOrder": None,
        "ordersWithError": set()
    }

    for plug, instance in logic.Iterator(plugins, context):
        state["nextOrder"] = plug.order

        if test(**state):
            log.error("Stopped due to: %s" % test(**state))
            break

        try:
            result = plugin.process(plug, context, instance)

        except:
            # This exception is unexpected
            log.error("An exception occurred.\n")
            raise

        else:
            # Make note of the order at which the
            # potential error error occured.
            if result["error"]:
                state["ordersWithError"].add(plug.order)

        if isinstance(result, Exception):
            log.error("An unexpected error happened: %s" % result)
            break

        error = result["error"]
        if error is not None:
            print(error)

    return context


def select(*args, **kwargs):
    """Convenience function for selection"""
    return _convenience(0.5, *args, **kwargs)


def validate(*args, **kwargs):
    """Convenience function for validation"""
    return _convenience(1.5, *args, **kwargs)


def extract(*args, **kwargs):
    """Convenience function for extraction"""
    return _convenience(2.5, *args, **kwargs)


def conform(*args, **kwargs):
    """Convenience function for conform"""
    return _convenience(3.5, *args, **kwargs)


collect = select
integrate = conform
run = publish  # Alias


def _convenience(order, *args, **kwargs):
    plugins = [p for p in plugin.discover()
               if p.order < order]

    args = list(args)
    if len(args) > 1:
        args[1] = plugins
    else:
        kwargs["plugins"] = plugins
    return publish(*args, **kwargs)


# Backwards compatibility
def publish_all(*args, **kwargs):
    warnings.warn("pyblish.util.publish_all has been "
                  "deprecated; use publish()")
    return publish(*args, **kwargs)


def validate_all(*args, **kwargs):
    warnings.warn("pyblish.util.validate_all has been "
                  "deprecated; use validate()")
    return validate(*args, **kwargs)




# Standard library
import os

# Local library
import pyblish.lib
import pyblish.plugin

from . import lib
from nose.tools import (
    with_setup,
    raises
)


package_path = pyblish.lib.main_package_path()
plugin_path = os.path.join(package_path, 'tests', 'plugins')

pyblish.plugin.deregister_all_paths()
pyblish.plugin.register_plugin_path(plugin_path)


@with_setup(lib.setup, lib.teardown)
def test_data():
    """The data() interface works"""

    ctx = pyblish.plugin.Context()

    # Not passing a key returns all data as a dict,
    # but there is none yet.
    assert ctx.data(key=None) == dict()

    key = 'test_key'

    ctx.set_data(key=key, value=True)
    assert ctx.data(key=key) is True
    assert ctx.has_data(key=key) is True
    ctx.remove_data(key=key)
    assert ctx.data(key=key) is None
    assert ctx.has_data(key=key) is False


@with_setup(lib.setup, lib.teardown)
def test_add_remove_instances():
    """Adding instances to context works"""
    ctx = pyblish.plugin.Context()
    inst = pyblish.plugin.Instance(name='Test', parent=ctx)
    ctx.remove(inst)


@with_setup(lib.setup, lib.teardown)
def test_instance_equality():
    """Instance equality works"""
    inst1 = pyblish.plugin.Instance('Test1')
    inst2 = pyblish.plugin.Instance('Test2')
    inst3 = pyblish.plugin.Instance('Test2')

    assert inst1 != inst2
    assert inst2 == inst3


def test_context_itemgetter():
    """Context.get() works"""
    context = pyblish.api.Context()
    context.create_instance("MyInstanceA")
    context.create_instance("MyInstanceB")

    assert context["MyInstanceA"].name == "MyInstanceA"
    assert context["MyInstanceB"].name == "MyInstanceB"
    assert context.get("MyInstanceA").name == "MyInstanceA"
    assert context.get("MyInstanceB").name == "MyInstanceB"
    assert context[0].name == "MyInstanceA"
    assert context[1].name == "MyInstanceB"


def test_in():
    """Querying whether an Instance is in a Context works"""

    context = pyblish.api.Context()
    context.create_instance("MyInstance")
    assert "MyInstance" in context
    assert "NotExist" not in context


def test_add_to_context():
    """Adding to Context is deprecated, but still works"""
    context = pyblish.api.Context()
    instance = pyblish.api.Instance("MyInstance")
    context.add(instance)
    context.remove(instance)


@raises(KeyError)
def test_context_getitem_nonexisting():
    """Getting a nonexisting item from Context throws a KeyError"""

    context = pyblish.api.Context()
    context.create_instance("MyInstance")
    assert context.get("NotExist") is None
    context["NotExist"]


@raises(IndexError)
def test_context_getitem_outofrange():
    """Getting a item out of range throws an IndexError"""

    context = pyblish.api.Context()
    context.create_instance("MyInstance")
    context[10000]


def test_context_getitem_validrange():
    """Getting an existing item works well"""

    context = pyblish.api.Context()
    context.create_instance("MyInstance")
    assert context[0].data["name"] == "MyInstance"


if __name__ == '__main__':
    test_add_remove_instances()



"""Pyblish initialisation

Attributes:
    _registered_paths: Currently registered plug-in paths.
    _registered_plugins: Currently registered plug-ins.

"""

from .version import version, version_info, __version__


_registered_paths = list()
_registered_callbacks = dict()
_registered_plugins = dict()
_registered_services = dict()
_registered_test = dict()
_registered_hosts = list()


__all__ = [
    "version",
    "version_info",
    "__version__",
    "_registered_paths",
    "_registered_callbacks",
    "_registered_plugins",
    "_registered_services",
    "_registered_test",
    "_registered_hosts",
]



import os

import pyblish
import pyblish.cli
import pyblish.api

from pyblish.vendor.click.testing import CliRunner


def ctx():
    """Return current Click context"""
    return pyblish.cli._ctx


def context():
    """Return current context"""
    return ctx().obj["context"]


def test_visualise_environment_paths():
    """Visualising environment paths works well"""
    current_path = os.environ.get("PYBLISHPLUGINPATH")

    try:
        os.environ["PYBLISHPLUGINPATH"] = "/custom/path"

        runner = CliRunner()
        result = runner.invoke(pyblish.cli.main, ["--environment-paths"])

        assert result.output.startswith("/custom/path"), result.output

    finally:
        if current_path is not None:
            os.environ["PYBLISHPLUGINPATH"] = current_path


def test_publishing():
    """Basic publishing works"""

    count = {"#": 0}

    class Collector(pyblish.api.ContextPlugin):
        def process(self, context):
            self.log.warning("Running")
            count["#"] += 1
            context.create_instance("MyInstance")

    class Validator(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder

        def process(self, instance):
            count["#"] += 10
            assert instance.data["name"] == "MyInstance"
            count["#"] += 100

    pyblish.api.register_plugin(Collector)
    pyblish.api.register_plugin(Validator)

    runner = CliRunner()
    runner.invoke(pyblish.cli.main, ["publish"])

    assert count["#"] == 111, count



"""Compatibility module"""

import warnings
from . import plugin, lib

# Aliases
Selector = plugin.Collector
Conformer = plugin.Integrator


def sort(*args, **kwargs):
    warnings.warn("pyblish.api.sort has been deprecated; "
                  "use pyblish.api.sort_plugins")
    return plugin.sort(*args, **kwargs)


def deregister_all(*args, **kwargs):
    warnings.warn("pyblish.api.deregister_all has been deprecated; "
                  "use pyblish.api.deregister_all_paths")
    return plugin.deregister_all_paths(*args, **kwargs)


# AbstractEntity
#
# The below members represent backwards compatibility
# features, kept separate for maintainability as they
# will no longer be updated and eventually discarded.


@lib.deprecated
def set_data(self, key, value):
    """DEPRECATED - USE .data DICTIONARY DIRECTLY

    Modify/insert data into entity

    Arguments:
        key (str): Name of data to add
        value (object): Value of data to add

    """

    self.data[key] = value


@lib.deprecated
def remove_data(self, key):
    """DEPRECATED - USE .data DICTIONARY DIRECTLY

    Remove data from entity

    Arguments;
        key (str): Name of data to remove

    """

    self.data.pop(key)


@lib.deprecated
def has_data(self, key):
    """DEPRECATED - USE .data DICTIONARY DIRECTLY

    Check if entity has key

    Arguments:
        key (str): Key to check

    Return:
        True if it exists, False otherwise

    """

    return key in self.data


@lib.deprecated
def add(self, other):
    """DEPRECATED - USE .append

    Add member to self

    This is to mimic the interface of set()

    """

    return self.append(other)


@lib.deprecated
def remove(self, other):
    """DEPRECATED - USE .pop

    Remove member from self

    This is to mimic the interface of set()

    """

    index = self.index(other)
    return self.pop(index)


plugin.AbstractEntity.add = add
plugin.AbstractEntity.remove = remove
plugin.AbstractEntity.set_data = set_data
plugin.AbstractEntity.remove_data = remove_data
plugin.AbstractEntity.has_data = has_data


# Context

@lib.deprecated
def create_asset(self, *args, **kwargs):
    return self.create_instance(*args, **kwargs)


@lib.deprecated
def add(self, other):
    return super(plugin.Context, self).append(other)


plugin.Context.create_asset = create_asset
plugin.Context.add = add




import os
import pyblish.api


class CollectCurrentWorkingDirectory(pyblish.api.ContextPlugin):
    """Inject the current working directory into Context"""

    order = pyblish.api.CollectorOrder

    def process(self, context):
        context.data['cwd'] = os.getcwd()
